"""
util.py

Utility functions.
"""

from datetime import datetime
import glob
import hashlib
import json
import logging
from pathlib import Path
import os

from . import User, Group, Role
from seamm_dashboard import db

logger = logging.getLogger(__name__)

time_format = "%Y-%m-%d %H:%M:%S %Z"


def process_flowchart(flowchart_path):
    """Read in flowchart from file and process for addition to SEAMM datastore.

    Parameters
    ----------
        flowchart_path: str
            The location of the flowchart

    Returns
    -------
        flowchart_info: dict
            Dictionary of flowchart information.
    """
    flowchart_info = {}

    flowchart_info["path"] = os.path.abspath(flowchart_path)

    # Get the flowchart text
    with open(flowchart_path, "r") as f:
        flowchart_info["text"] = f.read()

    # Get the flowchart description
    with open(flowchart_path) as f:
        f.readline()
        f.readline()
        flowchart_info["json"] = json.load(f)

    # Get a hash of the flowchart contents
    flowchart_info["id"] = hashlib.md5(
        flowchart_info["text"].encode("utf-8")
    ).hexdigest()

    # Get the flowchart description.
    try:
        node0 = flowchart_info["json"]["nodes"][0]
        flowchart_info["description"] = node0["attributes"]["_description"]
    except KeyError:
        flowchart_info["description"] = "No description given."
    except Exception:
        flowchart_info["description"] = "The flowchart may be corrupted."

    return flowchart_info


def process_job(job_path):
    """Process path for adding job to datastore.

    Parameters
    ----------
        job_path: str
            Path to directory job is in.

    Returns
    -------
        job_info: dict
            The job information to be added to the datastore.
    """

    job_info = {"description": ""}

    # Look for a flowchart in the job path
    flow_path = os.path.join(job_path, "*.flow")
    flowchart = glob.glob(flow_path)

    if len(flowchart) > 1:
        raise ValueError("More than one flowchart found! Cannot add job.")

    if len(flowchart) < 1:
        # If no flowchart is found - not a job - can't store. Return None
        return None

    flowchart_info = process_flowchart(flowchart[0])

    # If there is a job_data.json file, extract data
    data_file = os.path.join(job_path, "job_data.json")
    if os.path.exists(data_file):
        try:
            with open(data_file, "r") as fd:
                data = json.load(fd)

            if "title" in data:
                job_info["title"] = data["title"]
            if "state" in data:
                job_info["status"] = data["state"]
            else:
                job_info["status"] = "unknown"
            if "start time" in data:
                job_info["submitted"] = datetime.strptime(
                    data["start time"], time_format
                )
                job_info["started"] = datetime.strptime(data["start time"], time_format)
            if "end time" in data:
                job_info["finished"] = datetime.strptime(data["end time"], time_format)
        except Exception as e:
            logger.warning("Encountered error reading job {}".format(job_path))
            logger.warning("Error: {}".format(e))

    job_info["flowchart_id"] = flowchart_info["id"]
    job_info["path"] = os.path.abspath(job_path)
    job_info["flowchart_path"] = flowchart_info["path"]

    # Attempt to read job ID from file path
    dir_name = os.path.basename(job_path)
    job_id = dir_name.split("_")[1].lstrip("0")
    job_info["id"] = int(job_id)

    return job_info


def file_owner(path):
    """Return the User object for the owner of a file or directory.

    The User is created if it does not exist.

    Parameters:
    -----------
    path : str or path
        The directory or file to check.

    Returns
    -------
    User object
    """

    item = Path(path)
    if item.exists():
        # Get the group first
        name = item.group()
        group = db.session.query(Group).filter_by(name=name).one_or_none()
        if group is None:
            group = Group(name=name)
            db.session.add(group)
            db.session.commit()

        # and now the user
        name = item.owner()
        user = db.session.query(User).filter_by(username=name).one_or_none()
        if user is None:
            admin_role = db.session.query(Role).filter_by(name="admin").one_or_none()

            if admin_role is None:
                admin_role = Role(name="admin")

            user = User(username=name, password="default", roles=[admin_role])
            user.groups.append(group)
            db.session.add(user)
            db.session.add(admin_role)
            db.session.add(group)
            db.session.commit()
        return user.id, group.id
    else:
        return None



from flask import request, render_template, flash, redirect, url_for

from flask_jwt_extended import jwt_required, get_current_user

from seamm_dashboard import db, authorize

from . import jobs

from seamm_dashboard.routes.jobs.forms import EditJob
from seamm_dashboard.models import Job


@jobs.route("/views/jobs/")
@jobs.route("/views//jobs/")
def jobs_list():
    return render_template("jobs/jobs_list.html", project=False)


@jobs.route("/views/jobs/<id>")
@jobs.route("/views//jobs/<id>")
@jwt_required(optional=True)
def job_details(id):
    job = Job.query.get(id)

    if not authorize.read(job):
        return render_template("401.html")

    edit_job = authorize.update(job)

    own_string = "You do not own this job."

    if job.owner == get_current_user():
        own_string = "You are the job owner."

    # Build the url ourselves.
    base_url = url_for("main.index")
    edit_url = base_url + f"jobs/{id}/edit"

    return render_template(
        "jobs/job_report.html",
        edit_job=edit_job,
        job=job,
        edit_url=edit_url,
        own_string=own_string,
    )


@jobs.route("/jobs/<job_id>/edit", methods=["GET", "POST"])
@jwt_required(optional=True)
def edit_job(job_id):

    job = Job.query.get(job_id)

    if not authorize.update(job):
        return render_template("401.html")

    form = EditJob()

    # Build the url ourselves.
    base_url = url_for("main.index")
    job_url = base_url + f"#jobs/{job_id}"

    if form.validate_on_submit():
        job.title = form.name.data
        job.description = form.notes.data
        db.session.commit()
        flash("Job updated successfully.", "successs")

        return redirect(job_url)

    elif request.method == "GET":
        form.name.data = job.title
        form.notes.data = job.description

    return render_template(
        "jobs/edit_job.html", title=f"Edit Job {job_id}", form=form, back_url=job_url
    )



from flask import render_template
from . import flowcharts


@flowcharts.route("/views/flowcharts")
def flowchart_list():
    return render_template("flowcharts/flowchart_list.html")


@flowcharts.route("/views/flowcharts/<id>")
@flowcharts.route("/views/flowcharts/<id>/<flowchart_keys>")
def flowchart_details(id):
    return render_template("flowcharts/render_flowchart.html")



"""Super short utility script to keep only the files from node_modules
that are used by the app.
"""

import re
import subprocess

import os
import shutil

if __name__ == "__main__":

    # manual addition of files which are needed but not picked up by the regex below.
    needed_but_unlisted = [
        os.path.join(
            "node_modules", "jstree", "dist", "themes", "default", "throbber.gif"
        ),
        os.path.join("node_modules", "jstree", "dist", "themes", "default", "32px.png"),
    ]

    full_path = os.path.abspath(os.path.join("..", "templates/"))

    bashCommand = f"grep -r {full_path} -e node_modules"

    with open("files.out", "w+") as f:
        process = subprocess.Popen(bashCommand.split(), stdout=f)
        output, error = process.communicate()

    with open("files.out", "r") as f:
        data = f.read()

    # Use regex to get files that are used in templates.
    pattern = re.compile("(node_modules/.+(?:css|js))")

    matches = pattern.findall(data)

    matches = list(set(matches))

    matches.extend(needed_but_unlisted)

    # assert False, matches

    tree = os.walk("node_modules")

    for dirpath, dirname, filename in tree:
        for f in filename:
            full_path = os.path.join(dirpath, f)

            if full_path not in matches:
                os.remove(full_path)

    # Second pass to delete empty folders
    tree = os.walk("node_modules", topdown=False)
    for dirpath, dirname, filename in tree:
        for name in dirname:
            folder_path = os.path.join(dirpath, name)

            if not os.listdir(folder_path):
                shutil.rmtree(folder_path)

    os.remove("files.out")



import types
import six

from sqlalchemy import Column, ForeignKey, Integer, or_
from sqlalchemy.orm import relationship, backref

from sqlalchemy.ext.declarative import declared_attr

from flask_authorize.mixins import PipedList, PermissionsMixin

from flask_authorize.plugin import (
    Authorizer,
    user_is_restricted,
    user_is_allowed,
    has_permission,
    user_has_role,
    user_in_group,
)

from flask import current_app

__all__ = ["generate_association_table", "AccessControlPermissionsMixin"]


def generate_association_table(
    entity_name, resource_name, entity_tablename=None, resource_tablename=None
):

    # Make them plural by adding 's' :)
    if not entity_tablename:
        entity_tablename = entity_name.lower() + "s"
    if not resource_tablename:
        resource_tablename = resource_name.lower() + "s"

    @declared_attr
    def left_id(cls):
        return Column(Integer, ForeignKey(f"{entity_tablename}.id"), primary_key=True)

    @declared_attr
    def right_id(cls):
        return Column(Integer, ForeignKey(f"{resource_tablename}.id"), primary_key=True)

    @declared_attr
    def entity_relationship(cls):
        return relationship(
            f"{entity_name}",
            backref=backref(f"special_{resource_tablename}", lazy="dynamic"),
        )

    @declared_attr
    def resource_relationship(cls):
        return relationship(
            f"{resource_name}",
            backref=backref(f"special_{entity_tablename}", lazy="dynamic"),
        )

    class PermissionsAssociationMixin:
        __tablename__ = f"{entity_tablename}_{resource_tablename}_association"

        entity_id = left_id
        resource_id = right_id

        locals()[f"{entity_tablename}"] = entity_relationship
        locals()[f"{resource_tablename}"] = resource_relationship

        permissions = Column(PipedList)

    return PermissionsAssociationMixin


class AccessControlPermissionsMixin(PermissionsMixin):
    @classmethod
    def authorized(cls, check):
        from flask_authorize.plugin import CURRENT_USER

        current_user = CURRENT_USER()

        ret = super().authorized(check)

        clauses = []

        # Check user for special permissions
        if hasattr(current_user, f"special_{cls.__tablename__}"):
            clauses.append(
                cls.id.in_(
                    [
                        x.resource_id
                        for x in getattr(current_user, f"special_{cls.__tablename__}")
                        if check in x.permissions
                    ]
                )
            )

        # Check if user is part of a group that has special access
        if current_user and hasattr(cls, "special_groups"):
            group_list = [
                x.special_groups.all()
                for x in cls.query.all()
                if x.special_groups.all()
            ]
            overlapping_groups = [
                y.resource_id
                for x in group_list
                for y in x
                if y.entity_id in [n.id for n in current_user.groups]
                if check in y.permissions
            ]
            clauses.append(cls.id.in_(overlapping_groups))

        return or_(ret, or_(*clauses))


def allowed(self, *args, **kwargs):
    from flask_authorize.plugin import CURRENT_USER

    # look to flask-login for current user
    user = kwargs.get("user")
    if user is None:
        user = CURRENT_USER()

    # otherwise, use current user method
    elif isinstance(user, types.FunctionType):
        user = user()

    # don't allow anything for anonymous users
    if user is None:
        if not current_app.config["AUTHORIZE_ALLOW_ANONYMOUS_ACTIONS"]:
            return False

    # authorize if user has relevant role
    if len(self.has_role):
        if user_has_role(user, self.has_role):
            return True
        elif not len(self.permission) and not len(self.create):
            return False

    # authorize if user has relevant group
    if len(self.in_group):
        if user_in_group(user, self.in_group):
            return True
        elif not len(self.permission) and not len(self.create):
            return False

    # authorize create privileges based on access
    if len(self.create):
        for model in self.create:
            if user_is_restricted(user, ["create"], model) or not user_is_allowed(
                user, ["create"], model
            ):
                return False

    # return if no additional permission check needed
    if len(self.permission) == 0:
        return True

    # check permissions on individual instances - all objects
    # must have authorization to proceed.
    operation = set(self.permission)
    for arg in args:

        if not isinstance(arg.__class__, six.class_types):
            continue

        # check role restrictions/allowances
        if user_is_restricted(user, operation, arg):
            return False

        if not user_is_allowed(user, operation, arg):
            return False

        # only check permissions for items that have set permissions
        check = current_app.config["AUTHORIZE_IGNORE_PROPERTY"]
        if hasattr(arg, check) and not getattr(arg, check):
            continue
        if not hasattr(arg, "permissions"):
            continue

        # check other permissions
        check = arg.permissions.get("other", [])
        permitted = has_permission(operation, check)

        # check user permissions
        if hasattr(arg, "owner"):
            if arg.owner == user:
                check = arg.permissions.get("owner", [])
                permitted |= has_permission(operation, check)

        # check group permissions
        if hasattr(arg, "group"):
            if hasattr(user, "groups"):
                if arg.group in user.groups:
                    check = arg.permissions.get("group", [])
                    permitted |= has_permission(operation, check)

        # check special permissions
        # should probably figure this out instead of cheating.
        operation = list(operation)[0]

        if hasattr(arg, "special_users") and CURRENT_USER():
            for assoc in arg.special_users.all():
                if CURRENT_USER().id == assoc.entity_id:
                    permitted |= operation in assoc.permissions

        if hasattr(arg, "special_groups") and CURRENT_USER():
            for group in arg.special_groups.all():
                if group.entity_id in [x.id for x in CURRENT_USER().groups]:
                    permitted |= operation in group.permissions

        if not permitted:
            return False

    return True


Authorizer.allowed = allowed



"""
Filetrs for Jinia templates
"""

from flask import current_app


def replace_empty(s):
    """Replace empty strings"""

    if not s:
        return current_app.config["REPLACE_NONE"]

    return s



"""
Table models for SEAMM datastore SQLAlchemy database.
"""

from datetime import datetime
from werkzeug.security import generate_password_hash, check_password_hash
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema
from marshmallow_sqlalchemy.fields import Related, Nested

# Patched flask authorize
from seamm_dashboard.flask_authorize_patch import (
    AccessControlPermissionsMixin,
    generate_association_table,
)

from seamm_dashboard import db, jwt

#############################
#
# SQLAlchemy Models
#
#############################

# Authentication Mapping Tables for Access Control
UserJobMixin = generate_association_table("User", "Job")
UserFlowchartMixin = generate_association_table("User", "Flowchart")
UserProjectMixin = generate_association_table("User", "Project")
GroupJobMixin = generate_association_table("Group", "Job")
GroupProjectMixin = generate_association_table("Group", "Project")
GroupFlowchartMixin = generate_association_table("Group", "Flowchart")


class UserJobAssociation(db.Model, UserJobMixin):
    pass


class UserFlowchartAssociation(db.Model, UserFlowchartMixin):
    pass


class UserProjectAssociation(db.Model, UserProjectMixin):
    def __setattr__(self, name, value):
        """
        Change behavior of set attribute so that when a user gets permissions for a
        project, they get updated permissions for all jobs and flowcharts within the
        project.
        """
        from seamm_dashboard import db

        if name == "permissions":
            # See if there is an asociation between the group and project
            project = Project.query.filter_by(id=self.resource_id).one()

            if project.jobs:
                for job in project.jobs:
                    assoc = UserJobAssociation.query.filter_by(
                        entity_id=self.entity_id, resource_id=job.id
                    ).one_or_none()

                    if assoc:
                        assoc.permissions = value
                    else:
                        assoc = UserJobAssociation(
                            entity_id=self.entity_id,
                            resource_id=job.id,
                            permissions=value,
                        )

                    db.session.add(assoc)
                    db.session.commit()

            if project.flowcharts:
                for flowchart in project.flowcharts:
                    assoc = UserFlowchartAssociation.query.filter_by(
                        entity_id=self.entity_id, resource_id=flowchart.id
                    ).one_or_none()
                    if assoc:
                        assoc.permissions = value
                    else:
                        assoc = UserFlowchartAssociation(
                            entity_id=self.entity_id,
                            resource_id=flowchart.id,
                            permissions=value,
                        )

                    db.session.add(assoc)
                    db.session.commit()

        super().__setattr__(name, value)


class GroupJobAssociation(db.Model, GroupJobMixin):
    pass


class GroupProjectAssociation(db.Model, GroupProjectMixin):
    def __setattr__(self, name, value):
        """
        Change behavior of set attribute so that when a group gets permissions for a
        project, they get updated permissions for all jobs and flowcharts within the
        project.
        """
        from seamm_dashboard import db

        if name == "permissions":
            # See if there is an asociation between the group and project
            project = Project.query.filter_by(id=self.resource_id).one()

            if project.jobs:
                for job in project.jobs:
                    assoc = GroupJobAssociation.query.filter_by(
                        entity_id=self.entity_id, resource_id=job.id
                    ).one_or_none()
                    if assoc:
                        assoc.permissions = value
                    else:
                        assoc = GroupJobAssociation(
                            entity_id=self.entity_id,
                            resource_id=job.id,
                            permissions=value,
                        )

                    db.session.add(assoc)
                    db.session.commit()

            if project.flowcharts:
                for flowchart in project.flowcharts:
                    assoc = GroupFlowchartAssociation.query.filter_by(
                        entity_id=self.entity_id, resource_id=flowchart.id
                    ).one_or_none()
                    if assoc:
                        assoc.permissions = value
                    else:
                        assoc = GroupFlowchartAssociation(
                            entity_id=self.entity_id,
                            resource_id=flowchart.id,
                            permissions=value,
                        )

                    db.session.add(assoc)
                    db.session.commit()

        super().__setattr__(name, value)


class GroupFlowchartAssociation(db.Model, GroupFlowchartMixin):
    pass


user_group = db.Table(
    "user_group",
    db.Column("user", db.Integer, db.ForeignKey("users.id"), primary_key=True),
    db.Column("group", db.Integer, db.ForeignKey("groups.id"), primary_key=True),
)

user_role = db.Table(
    "user_role",
    db.Column("user", db.Integer, db.ForeignKey("users.id"), primary_key=True),
    db.Column("role", db.Integer, db.ForeignKey("roles.id"), primary_key=True),
)


flowchart_project = db.Table(
    "flowchart_project",
    db.Column(
        "flowchart", db.String(32), db.ForeignKey("flowcharts.id"), primary_key=True
    ),
    db.Column("project", db.Integer, db.ForeignKey("projects.id"), primary_key=True),
)

job_project = db.Table(
    "job_project",
    db.Column("job", db.Integer, db.ForeignKey("jobs.id"), primary_key=True),
    db.Column("project", db.Integer, db.ForeignKey("projects.id"), primary_key=True),
)


class User(db.Model):
    __tablename__ = "users"

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String, unique=True, nullable=False)
    first_name = db.Column(db.String)
    last_name = db.Column(db.String)
    email = db.Column(db.String, unique=True)
    password_hash = db.Column(db.String)
    added = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    status = db.Column(db.String, default="active")

    roles = db.relationship("Role", secondary=user_role, back_populates="users")
    groups = db.relationship("Group", secondary=user_group, back_populates="users")

    @property
    def password(self):
        raise AttributeError("password is not a readable attribute")

    @password.setter
    def password(self, password):
        self.password_hash = generate_password_hash(password)

    def verify_password(self, password):
        return check_password_hash(self.password_hash, password)


@jwt.user_lookup_loader
def user_loader_callback(jwt_header, jwt_payload):
    """Function for app, to return user object"""

    if jwt_header:
        username = jwt_payload["sub"]["username"]
        user = User.query.filter_by(username=username).one_or_none()

        return user
    else:
        # return None / null
        return None


class Group(db.Model):
    __tablename__ = "groups"

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, unique=True, nullable=False)

    users = db.relationship("User", secondary=user_group, back_populates="groups")


class Role(db.Model):
    __tablename__ = "roles"

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, unique=True, nullable=False)

    users = db.relationship("User", secondary=user_role, back_populates="roles")


class Flowchart(db.Model, AccessControlPermissionsMixin):
    __tablename__ = "flowcharts"

    id = db.Column(db.String(32), nullable=False, primary_key=True)
    title = db.Column(db.String(100), nullable=True)
    description = db.Column(db.Text, nullable=True)
    path = db.Column(db.String, unique=True)
    text = db.Column(db.Text, nullable=False)
    json = db.Column(db.JSON, nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)

    jobs = db.relationship("Job", back_populates="flowchart", lazy=True)
    projects = db.relationship(
        "Project", secondary=flowchart_project, back_populates="flowcharts"
    )

    def __repr__(self):
        return f"Flowchart(id={self.id}, description={self.description}, path={self.path})"  # noqa: E501


class Job(db.Model, AccessControlPermissionsMixin):
    __tablename__ = "jobs"

    id = db.Column(db.Integer, primary_key=True)
    flowchart_id = db.Column(db.String(32), db.ForeignKey("flowcharts.id"))
    title = db.Column(db.String, nullable=True)
    description = db.Column(db.Text, nullable=True)
    path = db.Column(db.String, unique=True)
    submitted = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    started = db.Column(db.DateTime)
    finished = db.Column(db.DateTime, nullable=True)
    status = db.Column(db.String, nullable=False, default="imported")

    flowchart = db.relationship("Flowchart", back_populates="jobs")
    projects = db.relationship("Project", secondary=job_project, back_populates="jobs")

    def __repr__(self):
        return f"Job(path={self.path}, flowchart_id={self.flowchart}, submitted={self.submitted})"  # noqa: E501


class Project(db.Model, AccessControlPermissionsMixin):
    __tablename__ = "projects"

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, nullable=False, unique=True)
    description = db.Column(db.String(1000), nullable=True)
    path = db.Column(db.String, unique=True)

    flowcharts = db.relationship(
        "Flowchart", secondary=flowchart_project, back_populates="projects"
    )
    jobs = db.relationship("Job", secondary=job_project, back_populates="projects")

    def __repr__(self):
        return f"Project(name={self.name}, path={self.path}, description={self.description})"  # noqa: E501

    def set_permissions(self, permissions):
        super().set_permissions(permissions)

        for job in self.jobs:
            job.set_permissions(permissions)
            db.session.add(job)
            db.session.commit()

        for flowchart in self.flowcharts:
            flowchart.set_permissions(permissions)
            db.session.add(flowchart)
            db.session.commit()


#############################
#
# Marshmallow
#
#############################


class FlowchartSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Flowchart
        exclude = (
            "json",
            "text",
            "owner_permissions",
            "group_permissions",
            "other_permissions",
        )

    owner = Related("username")
    group = Related("name")


class ProjectSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Project
        exclude = ("owner_permissions", "group_permissions", "other_permissions")

    owner = Related("username")
    group = Related("name")


class JobSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Job
        exclude = (
            "flowchart",
            "owner_permissions",
            "group_permissions",
            "other_permissions",
        )

    owner = Related("username")
    group = Related("name")
    projects = Nested(
        ProjectSchema(
            only=(
                "name",
                "id",
            ),
            many=True,
        )
    )


class UserSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = User
        exclude = ("password_hash",)


class GroupSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Group

    users = Nested(
        UserSchema(
            only=(
                "username",
                "id",
            ),
            many=True,
        )
    )


class RoleSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Role



from flask_wtf import FlaskForm
from wtforms import StringField, TextAreaField, SubmitField


class EditJob(FlaskForm):
    name = StringField("Job Title")
    notes = TextAreaField("Description")
    submit = SubmitField("Update Job")
